var tryit$ = (function () {
	'use strict';

	function asHTML(x) {
		return x.replace(/&/g, '~AMP~').replace(/</g,'&lt;').replace(/>/g,'&gt;').replace(/~AMP~/g,'&amp;');
	}

	function H(s) {
		return ({_toHtml: () => '<br/><p><b>' + asHTML(s) + '</b></p>'});
	}

	let WINDOW_LOCATION = window.location.pathname;
	const EMPTY_ELEMENT = { innerText: '' };

	function Identity(x) { return x; }

	function isEmptyElement(x) { 
		return x === null || x === EMPTY_ELEMENT; 
	}

	function $e(name) {
		if(typeof name !== 'string') return name;
		var e = document.getElementById(name);
		if (!e) return EMPTY_ELEMENT;
		return e;
	}

	function qs(sel,base) {
		return (typeof sel === 'string')?(base||document).querySelector(sel):sel;
	}

	function qsA(arg,base) {
		return (base||document).querySelectorAll(arg);
	}

	function json(x) {
		return JSON.stringify(x,null,' ');
	}

	function isTag(elem, tagName) {
		if(!elem || !elem.tagName ) return false;
		return elem.tagName.toLowerCase() === tagName.toLowerCase();
	}

	function dataset(elem) {
		if(!elem || !elem.dataset) return ({});
		return elem.dataset;
	}

	function objInfo(c) {
		const instanceMethods = Object.getOwnPropertyNames(c.prototype)
			.filter(prop => prop != 'constructor');
		//console.log(instanceOnly);
		const staticMethods = Object.getOwnPropertyNames(c)
			.filter(prop => typeof c[prop] === 'function');
		//console.log(staticOnly);
		return {instanceMethods, staticMethods};
	}

	/**
	 * Convert and array-like into an Array, if it
	 * @param  {any} arrayLike An array like ojject or an array
	 * @return {[any]}           an array
	 */
	function asArray(arrayLike) {
		if(arrayLike === undefined || arrayLike === null) return [];
		if(typeof arrayLike === 'string') return [arrayLike];
		if(Array.isArray(arrayLike)) return arrayLike;

		if( arrayLike instanceof NodeList  ) {
			return Array.prototype.slice.call(arrayLike,0);
		}
		else if( arrayLike instanceof Map) {
			return Array.from(arrayLike.entries());
		}
		else if(arrayLike instanceof Set) {
			return Array.from(arrayLike);
		} else if(typeof arrayLike.forEach === 'function') {
			let res = [];
			arrayLike.forEach(n => res.push(n));
			return res;
		}

		if(!arrayLike || arrayLike.length === undefined ) {
			return [arrayLike];
		}

		let res = [];
		for(let i=0; i< arrayLike.length; i++) res.push(arrayLike[i]);
		return res;
	}

	function _addRemoveCSSclass(next_button,classToRemove,classToAdd) {
		if(next_button) {
			let b = $e(next_button);
			if(!isEmptyElement(b)) {
				if(classToRemove) asArray(classToRemove).forEach(cls => b.classList.remove(cls));
				if(classToAdd) asArray(classToAdd).forEach(cls => b.classList.add(cls));
				return b;
			}
		}
		return ({});
	}

	function addRemoveCSSclass(next_button,classToAdd, classToRemove) {
		return _addRemoveCSSclass(next_button+'-run', classToAdd, classToRemove);
	}

	let editorFor = {};
	//export let allEditors;


	let __editorsPending = []; 
	function getPendingEditors() {
		return __editorsPending.slice();
	}

	function setPendingEditors(list) {
		__editorsPending = list || [];
		return list;
	}

	/**
	 * [getTocLookup description]
	 * @return {[type]} [description]
	 */

	function getTocLookup() {

		let lookup = {};
		let m = mapper();
		let pageList = m.map( ({page, children}) => [page,children[0]] );
		let values = flatten(m.map(({children}) => children));
		let current = '';
		//return pages;
		values.forEach(v => {
			if(v[0] == '*') {
				current = v.substr(1);
				lookup[current] = current;
			} else {
				let [, id] = v.split(':');
				lookup[id] = current;
			}
		});

		pageList.forEach(([p,h1]) => lookup[p] = h1.substr(1));
		return lookup;

		// =============================
		function asArray(arr) {
			return Array.prototype.slice.call(arr); 
		}

		function pages() { return asArray(qsA('.try-page')); }

		function hFlatten(e) {
			if(e.children && e.children.length === 0) return type(e);
			else return [type(e), ...asArray(e.children).map(hFlatten)];
		}

		function type(e) {
			if(!e) return undefined;
			//console.log(e.tagName);
			if(!e.id) return '';
			let ty = e.tagName;
			if(ty.match(/^h\d/i)) return '*'+e.id;
			if(ty === 'A' ) return ty+':'+e.id;
			return '';
		}

		function mapper() { 
			return pages()
				.map(
					e => ({
						page:e.id , 
						children: flatten(asArray(e.children)
							.flatMap(hFlatten))
							.filter(Identity)
					})
				); 
		}

		function flatten(arr) {
			if(!Array.isArray(arr)) return arr;
			if(!arr.some(Array.isArray)) return arr;
			return [].concat(...arr.map(flatten));
		}
	}

	function findSegment(elem) {
		if(elem === undefined) {
			let segment =qs('div[data-pagevisible="true"]'); 
			return (segment||{});
		}

		if( dataset(elem).pagevisible ) return elem;
		let segment = elem.closest('div[data-pagevisible]');
		return (segment || {});

	}

	/**
	 *  Proxy for CodeMirror as a place-holder until the CodeMirror editor has
	 *  been created
	 */
	class EditorProxy {
		/**
		 * [constructor description]
		 * @param  {[type]} name [description]
		 * @return {[type]}      [description]
		 */
		constructor(name) {
			this.name = name;
			this._editor = undefined;
			this.requiredContent = undefined;
			this.reqOptions = new Map();
		}
		
		/**
		 * [hasEditor description]
		 * @return {Boolean} [description]
		 */
		hasEditor() {
			return this._editor !== undefined;
		}

		get editor() {
			return this._editor;
		}

		set editor(anEditor) {
			this._editor = anEditor;
			if(this.requiredContent) {
				anEditor.setValue(this.requiredContent);
				this.requiredContent = undefined;
			}
			if(this.reqOptions.length) {
				this.reqOptions.forEach((key,value) => anEditor.setOption(key, value));
				this.reqOptions = {};
			}
		}
		/**
		 * [getOption description]
		 * @param  {[type]} key [description]
		 * @return {[type]}     [description]
		 */
		getOption(key) {
			if(this.editor) return this.editor.getOption(key);
			return this.reqOptions.get(key);
		}
		/**
		 * [setOption description]
		 * @param {[type]} key   [description]
		 * @param {[type]} value [description]
		 */
		setOption(key,value) {
			if(this.editor) return this.editor.setOption(key,value);
			return this.reqOptions.set(key, value);
		}
		/**
		 * [getValue description]
		 * @return {[type]} [description]
		 */
		getValue() {
			if(this._editor) return this._editor.getValue('\n');
			return document.getElementById(this.name).value; 
		}
		/**
		 * [setValue description]
		 * @param {[type]} content [description]
		 */
		setValue(content) {
			if(this._editor) this._editor.setValue(content);
			else this.requiredContent = content; 
		}
		/**
		 * [isClean description]
		 * @param  {[type]}  val [description]
		 * @return {Boolean}     [description]
		 */
		isClean(val) {
			return this._editor === undefined || this._editor.isClean(val);
		}
		/**
		 * [toString description]
		 * @return {[type]} [description]
		 */
		toString() {
			return `Editor(${this.name})`;
		}
	}

	let editorData;

	function getEditorData() {
		return editorData;
	}

	function setEditorData(v) {
		editorData = v;
		return editorData;
	}

	setEditorData(( () => { // done
		let data = window.localStorage[WINDOW_LOCATION];
		if(data) {
			try {
				let obj = JSON.parse(data);
				let keys = Object.keys(obj);
				keys.forEach(k =>{
					let v = obj[k];
					if(typeof v === 'string') {
						obj[k] = { key: k, hash: sha1(v), content: v};
					}
				});
				return obj;
			} catch(e) {
				return ({});
			}
		}
		return ({});
	})());

	function getSavedContent(id) {
		const editorData = getEditorData();
		let saved = editorData[id];
		let originalContents = $e(id).value;
		let hash = window.sha1(originalContents);
		if(saved && saved.key === id && saved.hash === hash ) {
			return saved.content;
		}
		else {
			let res = Object.keys(editorData)
				.map( k => editorData[k])
				.find(saved => saved.hash === hash);
			return res? res.content : originalContents;
		}
	}

	function setEditorHeight(editor,lines) {
		let height = '';
		if(lines < 5 ) height = '5rem';
		else if( lines > 20 ) height = '40rem';
		else height = (lines*1.8)+'rem';
		editor.setSize('inherit', height);
		return editor;
	}

	//import { tryIt } from '../execute/_runAll';

	let tryIt; // greak circular reference
	function setupMakeEditor(doTryIt) {
		tryIt = doTryIt;
	}
	function _makeEditor(id, callback=Identity) {
		try {
			// let originalContents = textarea.value;
			// let contents = originalContents;
			// if(editorData[id]) {
			//   contents = editorData[id].content;
			// } else {
			//   editorData[id] = contents;
			// }
			let textarea = document.querySelector(`#${id}`);

			let contents = getSavedContent(id);
			const lines = contents.split('\n').length;
			const original = textarea.value;
			const theme = (original === contents)?tryit$colors.original:tryit$colors.saved;
			const editor = CodeMirror.fromTextArea(textarea, {
				lineNumbers: true,
				// mode: "javascript",
				//mode: "jsx",
				theme: theme,//"cobalt",
				matchBrackets: true,
				autoCloseBrackets: '()[]{}\'\'""``', 
				continueComments: 'Enter',
				extraKeys: {
					'Ctrl-Enter': execCode,
					'Cmd-Enter': execCode,
					'Ctrl-/': 'toggleComment',
					'Ctrl-F': 'search',
					'Ctrl-Space': 'autocomplete',

				},
				tabSize: 2,
				keyMap: 'sublime',
				lineWrapping: true,
				lint: { bitwise: true, esversion: 10, notypeof: true, expr: true, asi: true },
				foldGutter: true,
				gutters: ['CodeMirror-lint-markers','CodeMirror-linenumbers', 'CodeMirror-foldgutter'],
				indentUnit: 2,
				mode: {name: 'javascript', globalVars: true}

			});
			if(original !== contents) {
				editor.setValue(contents);
				editor.markClean();
			}
			editor.tryitState = theme;
			editor.on('change', () => editorChanged(editor));
			// getPendingEditors().push(id);
			// __editors.push(id);
			// editorFor[id] = editor;
			setEditorHeight(editor,lines);
			callback(id, editor);
			function execCode() { return tryIt(id,editor);}

			// qs(`#${id}-run`).onclick = execCode;
			// function execCode() { return tryIt(id,editor);}
		}
		catch(err) {
			alert('Error creating editor ' +id+' ' +err.toString());
		}
	}

	function editorChanged(editor) {
		let theme = editor.getOption('theme');
		if(editor.isClean())
			editor.setOption('theme', editor.tryitState);
		else if(theme !== tryit$colors.edited ) {
			editor.setOption('theme', tryit$colors.edited);
		}
				
	}

	class PageInfo {
		constructor(pageList) {
			this.pageMap = new Map(asArray(pageList).map((e,ix) => [e.id, ix]));
			this.contents = new Map();
		}

		set(pageId, anEditorList) {
			this.contents.set(pageId, anEditorList);
		}

		pageIx(aPageName) {
			let ix = this.pageMap.get(aPageName);
			return ix === undefined ?-1:ix;
		}
		compare(page1,page2) {
			return this.pageIx(page1) - this.pageIx(page2);
		}

		showPage(pageId) {
			let editorList = this.contents.get(pageId);
			console.log('_makeEditor', _makeEditor);
			editorList.forEach(anEditor =>{
				if(!anEditor.hasEditor()) {
					_makeEditor(anEditor.name, (id,editor) => {
						editorFor[id] = editor;
						anEditor.editor = editor;
								
					});
				}
			});
		}
	}

	//import {tryIt} from '../execute/_runAll';

	let _pageInfo;
	let tryIt$1 = () => false;
	/**
	 * If the number of 'pages' in the HTML gets large and there lots of tryit scripts
	 * the loading of the page can get very slow. So instead we Create the code mirror
	 * editors for only the visible pages.
	 *
	 * Although the html file has all the pages, each page is in its own &lt;div&gt:
	 * only one page is visible at a time. The first time a page becomes visible, we use
	 * the EditorProxy to actually render the editor. Using an editor proxy, the rest of the
	 * code can be completely unaware that the CodeMirror editors do not exist on the hidden pages
	 * (div)
	 *
	 * 
	 * @return {object} - {pageInfo, allEditors}
	 */
	function getPageInfo() {
		if(_pageInfo) return _pageInfo;
		
		let list = qsA('.try-page');
		let pageInfo = new PageInfo(list), allEditors = [];
		let pendingEd = getPendingEditors();

		list.forEach(p => {
			//console.log("Page", p.id);
			let editorsOnPage = [];
			pageInfo.set(p.id,editorsOnPage); // editorsOnPage is a little later in this function
			let editors = qsA('.tryit',p);
			editors.forEach(e => {
				let id = e.id;
				if(id) {
					let anEditorProxy = new EditorProxy(id);
					editorsOnPage.push(anEditorProxy);
					allEditors.push(anEditorProxy);
					pendingEd.push(id); setPendingEditors(pendingEd);
					editorFor[id] = anEditorProxy;
					qs(`#${id}-run`).onclick = ( ()=>tryIt$1(id,anEditorProxy) );
				}
			});

		});
		return (_pageInfo = {pageInfo, allEditors});
	}

	function setupPageInfo(doTryit) {
		tryIt$1 = doTryit;
	}

	/**
	 * [setDisplay description]
	 * @param {[type]} elem      [description]
	 * @param {[type]} type      [description]
	 * @param {[type]} otherElem [description]
	 */
	function setDisplay(elem, type, otherElem) {
		if(!elem || !elem.dataset) return;
		if(otherElem && type == 'false') {
			delete otherElem.style.height;
			if(otherElem.offsetTop > elem.offsetTop) {
				let pos = window.scrollY || window.screenTop; 
				window.scrollTo(0, pos-elem.offsetHeight);
			}
		}
		elem.dataset.pagevisible = type;
	//	 elem.style.display = (type==='false')?'none':'block'; // may nood to enable this
	}

	function makeSegmentVisible(elem, timeout=20) {
		return new Promise(resolve => window.requestAnimationFrame(()=> doit(resolve)));

		function doit(resolve) {
			let [curSeg, segment] = [undefined, elem].map(findSegment); // find
			if(curSeg !== segment) {
				let pos = window.scrollY||window.screenTop;
				setDisplay(segment, 'true');
				let {pageInfo} = getPageInfo();
				pageInfo.showPage(segment.id);
				if(pageInfo.compare(segment.id,curSeg.id)<0) {
					let height = segment.offsetHeight;
					window.scrollTo(0,pos+height);
				}
				else if(segment.offsetHeight < window.innerHeight-5){
					segment.style.height = Math.round(window.innerHeight+5)+'px';
				}
			}
			return setTimeout( () => resolve([segment, curSeg]),timeout);
		}
	}

	function easeIn(start, pos, end) {
		const abs = Math.abs;
		if(abs(start-pos) < 100 || abs(end-pos) < 100 ) return 2;
		const diff =  abs(start-end);
		if(diff > 5000 ) return 100;
		if(diff > 1000 ) return 30;
		if(diff > 500)   return 10;
		return 5;
	}

	function scrollToSmoothly(posFn, time,callback){
		/*Time is only applicable for scrolling upwards*/
		/*Code written by hev1*/
		/*pos is the y-position to scroll to (in pixels)*/
		let v = posFn;
		if(typeof posFn !== 'function') posFn = () => v;
		let pos = posFn();
		if(isNaN(pos)){
			throw 'Position must be a number';
		}
		if(pos<0){
			//throw "Position can not be negative";
			pos = 0;
		}
		let start = (window.scrollY||window.screenTop);
		let currentPos = start;
		if(currentPos<pos){
			let t = 10;
			for(let i = currentPos; i <= pos+15; i+=10){
				t+=10;
				let v = i;
				setTimeout(() => {
					window.scrollTo(0, v);
					pos = posFn();
				}, t/2);
			}
			if(callback) {
				setTimeout(() => callback(), t/2 + 50);
			}
		} else {
			time = time || 2;
			let i = currentPos;
			let x;
			x = setInterval(() =>
			{
				window.scrollTo(0, i);
				pos = posFn();
				let delta = easeIn(start,i,pos);
				if(delta < 0) delta = -delta;
				i -= delta;
				if(i<=pos){
					clearInterval(x);
					if(callback) callback();
				}
			}, time);
		}
	}

	function _jumpTag(aTag, toToggle=true) {
		jumpTag(aTag);
		let toggle = window.toggle || Identity;
		toToggle && toggle();
	}

	function jumpback() {
		if(getPendingEditors().length) jump(getPendingEditors()[0]);
	}

	function jump(h) {
		jumpTag('_'+h,70);
	}



	function jumpBack() {
		jumpback();
		let toggle = window.toggle || Identity;
		toggle();
	}

	let LAST_TARGET;

	let tocLookup; 
	function jumpTag(h,OFFSET,callback, noPush) {
		if(!tocLookup) tocLookup = getTocLookup();
		OFFSET = +(OFFSET||30);
		callback = callback || Identity;
		let elem = $e(h);
		if(LAST_TARGET === elem.id) return;
		//let [targetSeg, curSeg] = makeSegmentVisible(elem);
		let displayPromise = makeSegmentVisible(elem,OFFSET);
		//pageInfo.showPage(targetSeg.id);
		//if(targetSeg !== curSeg) setDisplay(curSeg, 'false');
		displayPromise.then(([targetSeg, curSeg]) => {
			//const lastsScoll = () => elem.scrollIntoView({behavior: "smooth", block: "start"});
			const lastsScoll = () => scrollToSmoothly(toHeader(elem).offsetTop-OFFSET, 10);
			scrollToSmoothly(elem.offsetTop-OFFSET, 10,() => {
				try {
					callback(); 
					if(targetSeg !== curSeg) setDisplay(curSeg, 'false',targetSeg);
					lastsScoll(); 
					LAST_TARGET = elem.id;
					if(!noPush) history.pushState(null,null,'#'+elem.id);
					let tocSel = tocLookup[elem.id];
					if(tocSel) {
						let tocElem = $e('toc_'+tocSel);
						let prev = qs('.toc.select');
						if(prev) _addRemoveCSSclass(prev, ['select'],[]);
						if(tocElem) _addRemoveCSSclass(tocElem, [],['select']);
					}
								
				// location.hash = elem.id;
				} catch (e) {
					alert('error jumping to: '+h+ 'location');
				}
			});
		}, 300);
						
	}



	function toHeader(elem) {
		if(dataset(elem).pagevisible) {
			return elem.querySelector('h1');
		}
		return elem;
	}

	let NO_DISPLAY = false;
	function setNoDisplay(flag) { NO_DISPLAY = !!flag; }
	function getNoDisplay() { return NO_DISPLAY; }

	let _lastlyStack = [];

	function _lastly(onDisplay, fn) {
		//console.log("lastly", onDisplay,fn, NO_DISPLAY)
		if(typeof onDisplay === 'boolean' && typeof fn === 'function') {
			if(onDisplay && !getNoDisplay()) {
				_lastlyStack.push(fn); // only add the function if we are displaying
			}
		}
		else if(typeof onDisplay === 'function') {
			fn = onDisplay;
			_lastlyStack.push(onDisplay);
		}
	}



	function runLastly() {
		let list = _lastlyStack.slice();
		_lastlyStack = [];
		return Promise.all(list.map(valOrFunc)).then(valOrFunc).catch(err => alert('Error in  '+err.toString()));
	}

	function clearLastly() {
		_lastlyStack = [];
	}

	function valOrFunc(v) {
		try {
			return  typeof v === 'function'? v(getNoDisplay()) : v;
		} catch(err) {
			if(!getNoDisplay())alert(err);
		}
	}

	let _displayStack = [];


	function clearDisplay() { 
		_displayStack= []; 
	}

	function getDisplayStack() { 
		return _displayStack; 
	}

	function pushDisplay(s, type='h') {
		//if(NO_DISPLAY) return;
		if(!(s instanceof Promise) && s !== undefined) s = Promise.resolve(s);
		_displayStack.push([s,type]);
	}

	function _show(...list) {
		const pd = pushDisplay;
		if(list.length > 1) {
				
			pd('<div class="display-container">');
			//pushDisplay(list.map(v => `<div>${display(v)}</div>\n`),'h');
			list.forEach(v => {pd('<div>'); pd(v,'d'); pd('</div>'); });
			pd('</div>');
		}
		else {
			list.forEach(v => pd(v,'d'));
		}
	}

	/*
	Copyright (c) 2009 James Padolsey.  All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	   1. Redistributions of source code must retain the above copyright
		  notice, this list of conditions and the following disclaimer.

	   2. Redistributions in binary form must reproduce the above copyright
		  notice, this list of conditions and the following disclaimer in the
		  documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY James Padolsey ``AS IS'' AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL James Padolsey OR CONTRIBUTORS BE LIABLE
	FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
	DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
	LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
	OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.

	The views and conclusions contained in the software and documentation are
	those of the authors and should not be interpreted as representing official
	policies, either expressed or implied, of James Padolsey.

	 AUTHOR James Padolsey (http://james.padolsey.com)
	 VERSION 1.03.0
	 UPDATED 29-10-2011
	 CONTRIBUTORS
		David Waller
	    Benjamin Drucker

	*/

	var prettyPrint = (function prettyPrint(){
		
		/* These "util" functions are not part of the core
		   functionality but are  all necessary - mostly DOM helpers */
		
		var util = {
			
			el: function(type, attrs) {
				
				/* Create new element */
				var el = document.createElement(type), attr;
				
				/*Copy to single object */
				attrs = util.merge({}, attrs);
				
				/* Add attributes to el */
				if (attrs && attrs.style) {
					//var styles = attrs.style;
					util.applyCSS( el, attrs.style );
					delete attrs.style;
				}
				for (attr in attrs) {
					if (attrs.hasOwnProperty(attr)) {
						el[attr] = attrs[attr];
					}
				}
				
				return el;
			
			},
			
			applyCSS: function(el, styles) {
				/* Applies CSS to a single element */
				for (var prop in styles) {
					if (styles.hasOwnProperty(prop)) {
						try{
							/* Yes, IE6 SUCKS! */
							el.style[prop] = styles[prop];
						}catch(e){
							console.log('*');
						}
					}
				}
			},
			
			txt: function(t) {
				/* Create text node */
				return document.createTextNode(t);
			},
			
			row: function(cells, type, cellType) {
				
				/* Creates new <tr> */
				cellType = cellType || 'td';
				
				/* colSpan is calculated by length of null items in array */
				var colSpan = util.count(cells, null) + 1,
					tr = util.el('tr'), td,
					attrs = {
						style: util.getStyles(cellType, type),
						colSpan: colSpan,
						onmouseover: function() {
							var tds = this.parentNode.childNodes;
							util.forEach(tds, function(cell){
								if (cell.nodeName.toLowerCase() !== 'td') { return; }
								util.applyCSS(cell, util.getStyles('td_hover', type));
							});
						},
						onmouseout: function() {
							var tds = this.parentNode.childNodes;
							util.forEach(tds, function(cell){
								if (cell.nodeName.toLowerCase() !== 'td') { return; }
								util.applyCSS(cell, util.getStyles('td', type));
							});
						}
					};
					
				util.forEach(cells, function(cell){
					
					if (cell === null) { return; }
					/* Default cell type is <td> */
					td = util.el(cellType, attrs);
					
					if (cell.nodeType) {
						/* IsDomElement */
						td.appendChild(cell);
					} else {
						/* IsString */
						td.innerHTML = util.shorten(cell.toString());
					}
					
					tr.appendChild(td);
				});
				
				return tr;
			},
			
			hRow: function(cells, type){
				/* Return new <th> */
				return util.row(cells, type, 'th');
			},
			
			table: function(headings, type){
				
				headings = headings || [];
				
				/* Creates new table: */
				var attrs = {
						thead: {
							style:util.getStyles('thead',type)
						},
						tbody: {
							style:util.getStyles('tbody',type)
						},
						table: {
							style:util.getStyles('table',type)
						}
					},
					tbl = util.el('table', attrs.table),
					thead = util.el('thead', attrs.thead);
					util.el('colgroup');
					var tbody = util.el('tbody', attrs.tbody);
					
				if (headings.length) {
					// for(let i=0; i<headings.length; i++) {
					// 	console.log("headings", headings, i);
					// 	let tcolStyle = i=0?({"class":"pp_keycol"}):{};
					// 	tcolgroup.appendChild(util.el('col', tcolStyle))
					// }
					
					tbl.appendChild(thead);
					//thead.appendChild(tcolgroup);
					thead.appendChild( util.hRow(headings, type) );
				}
				tbl.appendChild(tbody);
				
				return {
					/* Facade for dealing with table/tbody
					   Actual table node is this.node: */
					node: tbl,
					tbody: tbody,
					thead: thead,
					appendChild: function(node) {
						this.tbody.appendChild(node);
					},
					addRow: function(cells, _type, cellType){
						this.appendChild(util.row.call(util, cells, (_type || type), cellType));
						return this;
					}
				};
			},
			
			shorten: function(str) {
				var max = 200;
				str = str.replace(/^\s\s*|\s\s*$|\n/g,'');
				return str.length > max ? (str.substring(0, max-1) + '...') : str;
			},
			
			htmlentities: function(str) {
				return str.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
			},
			
			merge: function(target, source) {
				
				/* Merges two (or more) objects,
				   giving the last one precedence */
				
				if ( typeof target !== 'object' ) {
					target = {};
				}
				
				for (var property in source) {
					
					if ( source.hasOwnProperty(property) ) {
						
						var sourceProperty = source[ property ];
						
						if ( typeof sourceProperty === 'object' ) {
							target[ property ] = util.merge( target[ property ], sourceProperty );
							continue;
						}
						
						target[ property ] = sourceProperty;
						
					}
					
				}
				
				for (var a = 2, l = arguments.length; a < l; a++) {
					util.merge(target, arguments[a]);
				}
				
				return target;
			},
			
			count: function(arr, item) {
				var count = 0;
				for (var i = 0, l = arr.length; i< l; i++) {
					if (arr[i] === item) {
						count++;
					}
				}
				return count;
			},
			
			thead: function(tbl) {
				return tbl.getElementsByTagName('thead')[0];
			},
			
			forEach: function(arr, max, fn) {
				
				if (!fn) {
					fn = max;
				}

				/* Helper: iteration */
				var len = arr.length,
					index = -1;
				
				while (++index < len) {
					if(fn( arr[index], index, arr ) === false) {
						break;
					}
				}
				
				return true;
			},
			
			type: function(v){
				try {
					/* Returns type, e.g. "string", "number", "array" etc.
					   Note, this is only used for precise typing. */
					if (v === null) { return 'null'; }
					if (v === undefined) { return 'undefined'; }
					var oType = Object.prototype.toString.call(v).match(/\s(.+?)\]/)[1].toLowerCase();
					if (v.nodeType) {
						if (v.nodeType === 1) {
							return 'domelement';
						}
						return 'domnode';
					}
					if (/^(string|number|array|regexp|function|date|boolean)$/.test(oType)) {
						return oType;
					}
					if (typeof v === 'object') {
						return v.jquery && typeof v.jquery === 'string' ? 'jquery' : 'object';
					}
					if (v === window || v === document) {
						return 'object';
					}
					return 'default';
				} catch(e) {
					return 'default';
				}
			},
			
			within: function(ref) {
				/* Check existence of a val within an object
				   RETURNS KEY */
				return {
					is: function(o) {
						for (var i in ref) {
							if (ref[i] === o) {
								return i;
							}
						}
						return '';
					}
				};
			},
			
			common: {
				circRef: function(obj, key, settings) {
					return util.expander(
						'[POINTS BACK TO <strong>' + (key) + '</strong>]',
						'Click to show this item anyway',
						function() {
							this.parentNode.appendChild( prettyPrintThis(obj,{maxDepth:1}) );
						}
					);
				},
				depthReached: function(obj, settings) {
					return util.expander(
						'[DEPTH REACHED]',
						'Click to show this item anyway',
						function() {
							try {
								this.parentNode.appendChild( prettyPrintThis(obj,{maxDepth:1}) );
							} catch(e) {
								this.parentNode.appendChild(
									util.table(['ERROR OCCURED DURING OBJECT RETRIEVAL'],'error').addRow([e.message]).node   
								);
							}
						}
					);
				}
			},
			
			getStyles: function(el, type) {
				type = prettyPrintThis.settings.styles[type] || {};
				return util.merge(
					{}, prettyPrintThis.settings.styles['default'][el], type[el]
				);
			},
			
			expander: function(text, title, clickFn) {
				return util.el('a', {
					innerHTML:  util.shorten(text) + ' <b style="visibility:hidden;">[+]</b>',
					title: title,
					onmouseover: function() {
						this.getElementsByTagName('b')[0].style.visibility = 'visible';
					},
					onmouseout: function() {
						this.getElementsByTagName('b')[0].style.visibility = 'hidden';
					},
					onclick: function() {
						this.style.display = 'none';
						clickFn.call(this);
						return false;
					},
					style: {
						cursor: 'pointer'
					}
				});
			},
			
			stringify: function(obj) {
				return JSON.stringify(obj);
			},
			_stringify: function(obj) {
				
				/* Bit of an ugly duckling!
				   - This fn returns an ATTEMPT at converting an object/array/anyType
					 into a string, kinda like a JSON-deParser
				   - This is used for when |settings.expanded === false| */
				
				var type = util.type(obj),
					str, first = true;
				if ( type === 'array' ) {
					str = '[';
					util.forEach(obj, function(item,i){
						str += (i===0?'':', ') + util.stringify(item);
					});
					return str + ']';
				}
				if (typeof obj === 'object') {
					str = '{';
					for (var i in obj){
						if (obj.hasOwnProperty(i)) {
							str += (first?'':', ') + i + ':' + util.stringify(obj[i]);
							first = false;
						}
					}
					return str + '}';
				}
				if (type === 'regexp') {
					return '/' + obj.source + '/';
				}
				if (type === 'string') {
					return '"' + obj.replace(/"/g,'\\"') + '"';
				}
				return obj.toString();
			},
			
			headerGradient: (function(){
				
				var canvas = document.createElement('canvas');
				if (!canvas.getContext) { return ''; }
				var cx = canvas.getContext('2d');
				canvas.height = 30;
				canvas.width = 1;
				
				var linearGrad = cx.createLinearGradient(0,0,0,30);
				linearGrad.addColorStop(0,'rgba(0,0,0,0)');
				linearGrad.addColorStop(1,'rgba(0,0,0,0.25)');
				
				cx.fillStyle = linearGrad;
				cx.fillRect(0,0,1,30);
				
				var dataURL = canvas.toDataURL && canvas.toDataURL();
				return 'url(' + (dataURL || '') + ')';
			
			})()
			
		};
		
		// Main..
		var prettyPrintThis = function(obj, options) {
			
			 /*
			 *	  obj :: Object to be printed					
			 *  options :: Options (merged with config)
			 */
			
			options = options || {};
			
			var settings = util.merge( {}, prettyPrintThis.config, options ),
				container = util.el('div'),
				currentDepth = 0,
				stack = {},
				hasRunOnce = false;
			
			/* Expose per-call settings.
			   Note: "config" is overwritten (where necessary) by options/"settings"
			   So, if you need to access/change *DEFAULT* settings then go via ".config" */
			prettyPrintThis.settings = settings;
			
			var typeDealer = {
				string : function(item){
					return util.txt('"' + util.shorten(item.replace(/"/g,'\\"')) + '"');
				},
				number : function(item) {
					return util.txt(item);
				},
				regexp : function(item) {
					
					var miniTable = util.table(['RegExp',null], 'regexp');
					var flags = util.table();
					var span = util.expander(
						'/' + item.source + '/',
						'Click to show more',
						function() {
							this.parentNode.appendChild(miniTable.node);
						}
					);
					
					flags
						.addRow(['g', item.global])
						.addRow(['i', item.ignoreCase])
						.addRow(['m', item.multiline]);
					
					miniTable
						.addRow(['source', '/' + item.source + '/'])
						.addRow(['flags', flags.node])
						.addRow(['lastIndex', item.lastIndex]);
						
					return settings.expanded ? miniTable.node : span;
				},
				domelement : function(element, depth) {
					
					var miniTable = util.table(['DOMElement',null], 'domelement'),
						props = ['id', 'className', 'innerHTML', 'src', 'href'], elname = element.nodeName || '';
					
					miniTable.addRow(['tag', '&lt;' + elname.toLowerCase() + '&gt;']);
						
					util.forEach(props, function(prop){
						if ( element[prop] ) {
							miniTable.addRow([ prop, util.htmlentities(element[prop]) ]);
						}
					});
					
					return settings.expanded ? miniTable.node : util.expander(
						'DOMElement (' + elname.toLowerCase() + ')',
						'Click to show more',
						function() {
							this.parentNode.appendChild(miniTable.node);
						}
					);
				},
				domnode : function(node){
					
					/* Deals with all DOMNodes that aren't elements (nodeType !== 1) */
					var miniTable = util.table(['DOMNode',null], 'domelement'),
						data =  util.htmlentities( (node.data || 'UNDEFINED').replace(/\n/g,'\\n') );
					miniTable
						.addRow(['nodeType', node.nodeType + ' (' + node.nodeName + ')'])
						.addRow(['data', data]);
					
					return settings.expanded ? miniTable.node : util.expander(
						'DOMNode',
						'Click to show more',
						function() {
							this.parentNode.appendChild(miniTable.node);
						}
					);
				},
				jquery : function(obj, depth, key) {
					return typeDealer['array'](obj, depth, key, true);
				},
				object : function(obj, depth, key) {
					
					/* Checking depth + circular refs */
					/* Note, check for circular refs before depth; just makes more sense */
					var stackKey = util.within(stack).is(obj);
					if ( stackKey ) {
						return util.common.circRef(obj, stackKey, settings);
					}
					stack[key||'TOP'] = obj;
					if (depth === settings.maxDepth) {
						return util.common.depthReached(obj, settings);
					}
					
					var table = util.table(['Object', null],'object'),
						isEmpty = true;
					
					for (var i in obj) {
						if (!obj.hasOwnProperty || obj.hasOwnProperty(i)) {
							var item = obj[i],
								type = util.type(item);
							isEmpty = false;
							try {
								table.addRow([i, typeDealer[ type ](item, depth+1, i)], type);
							} catch(e) {
								/* Security errors are thrown on certain Window/DOM properties */
								if (window.console && window.console.log) {
									console.log(e.message);
								}
							}
						}
					}
					
					if (isEmpty) {
						table.addRow(['<small>[empty]</small>']);
					} else {
						table.thead.appendChild(
							util.hRow(['key','value'], 'colHeader')
						);
					}
					
					var ret = (settings.expanded || hasRunOnce) ? table.node : util.expander(
						util.stringify(obj),
						'Click to show more',
						function() {
							this.parentNode.appendChild(table.node);
						}
					);
					
					hasRunOnce = true;
					
					return ret;
					
				},
				array : function(arr, depth, key, jquery) {
					
					/* Checking depth + circular refs */
					/* Note, check for circular refs before depth; just makes more sense */
					var stackKey = util.within(stack).is(arr);
					if ( stackKey ) {
						return util.common.circRef(arr, stackKey);
					}
					stack[key||'TOP'] = arr;
					if (depth === settings.maxDepth) {
						return util.common.depthReached(arr);
					}
					
					/* Accepts a table and modifies it */
					var me = jquery ? 'jQuery' : 'Array', table = util.table([me + '(' + arr.length + ')', null], jquery ? 'jquery' : me.toLowerCase()),
						isEmpty = true,
						count = 0;
					
					if (jquery){
						table.addRow(['selector',arr.selector]);
					}

					let LEN = arr.length;
					let LOW = LEN, HIGH = LEN;
					if( LEN >= 30 ) {
						LOW = 15;
						HIGH = LEN-15;
					}

					util.forEach(arr, function(item,i){
						if (settings.maxArray >= 0 && ++count > settings.maxArray) {
							table.addRow([
								i + '..' + (arr.length-1),
								typeDealer[ util.type(item) ]('...', depth+1, i)
							]);
							return false;
						}
						isEmpty = false;
						if(i < LOW || i >= HIGH)
							table.addRow([i, typeDealer[ util.type(item) ](item, depth+1, i)]);
						else if( i  === LOW ) {
							table.addRow([
								i + '..' + (HIGH-1),
								typeDealer[ util.type(item) ]('...', depth+1, i)
							]);
						}
					});

					if (!jquery){
						if (isEmpty) {
							table.addRow(['<small>[empty]</small>']);
						} else {
							table.thead.appendChild( util.hRow(['index','value'], 'colHeader') );
						}
					}
					
					return settings.expanded ? table.node : util.expander(
						util.stringify(arr),
						'Click to show more',
						function() {
							this.parentNode.appendChild(table.node);
						}
					);
					
				},
				'function' : function(fn, depth, key) {
					
					/* Checking JUST circular refs */
					var stackKey = util.within(stack).is(fn);
					if ( stackKey ) { return util.common.circRef(fn, stackKey); }
					stack[key||'TOP'] = fn;
					
					var miniTable = util.table(['Function',null], 'function');
						util.table(['Arguments']);
						var args = fn.toString().match(/\((.+?)\)/),
						body = fn.toString().match(/\(.*?\)\s+?\{?([\S\s]+)/)[1].replace(/\}?$/,'');
						
					miniTable
						.addRow(['arguments', args ? args[1].replace(/[^\w_,\s]/g,'') : '<small>[none/native]</small>'])
						.addRow(['body', body]);
						
					return settings.expanded ? miniTable.node : util.expander(
						'function(){...}',
						'Click to see more about this function.',
						function(){
							this.parentNode.appendChild(miniTable.node);
						}
					);
				},
				'date' : function(date) {
					
					var miniTable = util.table(['Date',null], 'date'),
						sDate = date.toString().split(/\s/);
					
					/* TODO: Make this work well in IE! */
					miniTable
						.addRow(['Time', sDate[4]])
						.addRow(['Date', sDate.slice(0,4).join('-')]);
						
					return settings.expanded ? miniTable.node : util.expander(
						'Date (timestamp): ' + (+date),
						'Click to see a little more info about this date',
						function() {
							this.parentNode.appendChild(miniTable.node);
						}
					);
					
				},
				'boolean' : function(bool) {
					return util.txt( bool.toString().toUpperCase() );
				},
				'undefined' : function() {
					return util.txt('UNDEFINED');
				},
				'null' : function() {
					return util.txt('NULL');
				},
				'default' : function() {
					/* When a type cannot be found */
					return util.txt('prettyPrint: TypeNotFound Error');
				}
			};
			
			container.appendChild( typeDealer[ (settings.forceObject) ? 'object' : util.type(obj) ](obj, currentDepth) );
			
			return container;
			
		};
		
		/* Configuration */
		
		/* All items can be overwridden by passing an
		   "options" object when calling prettyPrint */
		prettyPrintThis.config = {
			
			/* Try setting this to false to save space */
			expanded: true,
			
			forceObject: false,
			maxDepth: 3,
			maxArray: -1,  // default is unlimited
			styles: {
				array: {
					th: {
						backgroundColor: '#6DBD2A',
						color: 'white'
					}
				},
				'function': {
					th: {
						backgroundColor: '#D82525'
					}
				},
				regexp: {
					th: {
						backgroundColor: '#E2F3FB',
						color: '#000'
					}
				},
				object: {
					th: {
						backgroundColor: '#1F96CF'
					}
				},
				jquery : {
					th: {
						backgroundColor: '#FBF315'
					}
				},
				error: {
					th: {
						backgroundColor: 'red',
						color: 'yellow'
					}
				},
				domelement: {
					th: {
						backgroundColor: '#F3801E'
					}
				},
				date: {
					th: {
						backgroundColor: '#A725D8'
					}
				},
				colHeader: {
					th: {
						backgroundColor: '#EEE',
						color: '#000',
						textTransform: 'uppercase'
					}
				},
				'default': {
					table: {
						borderCollapse: 'collapse',
						width: '100%'
					},
					td: {
						padding: '5px',
						fontSize: '12px',
						backgroundColor: '#FFF',
						color: '#222',
						border: '1px solid #000',
						verticalAlign: 'top',
						fontFamily: '"Consolas","Lucida Console",Courier,mono',
						whiteSpace: 'nowrap'
					},
					td_hover: {
						/* Styles defined here will apply to all tr:hover > td,
							- Be aware that "inheritable" properties (e.g. fontWeight) WILL BE INHERITED */
					},
					th: {
						padding: '5px',
						fontSize: '12px',
						backgroundColor: '#222',
						color: '#EEE',
						textAlign: 'left',
						border: '1px solid #000',
						verticalAlign: 'top',
						fontFamily: '"Consolas","Lucida Console",Courier,mono',
						backgroundImage: util.headerGradient,
						backgroundRepeat: 'repeat-x'
					}
				}
			}
		};
		
		return prettyPrintThis;
		
	})();

	function display(d) {
		if(d && d._toHtml ) {
			return d._toHtml();
		}
		else if(isReactNode(d)) {
			let genID = 'RX'+Math.trunc(Math.random()*10000);
			_lastly(() => window.ReactDOM.render(d, document.getElementById(genID)));
			return `<div id="${genID}">ReactNode</div>`;
		}
		else if( d instanceof Set) {
			let setArrStr = Array.from(d).join(', ');

			return `<pre>Set{${asHTML(setArrStr)}}</pre>`;
		}
		else if( typeof d === 'string') {
			if( d && d.length > 20000) {
				d = d.substr(0,20000)+'... MORE';
				return '<pre>' + asHTML(d) + '</pre>';
			}
			//       if( d.length < 100) return  asHTML(d) + "<br/>"
			else if(d.indexOf('\n') === -1)
				return '<pre>' + JSON.stringify(asHTML(d)) + '</pre>';
			else
				return '<pre>' + (asHTML(d)) + '</pre>';
		}
		else if(isPrimitive(d)) return '<pre>' + d.toString()+ '</pre>';
		else if(smallArray(d)) {
			let v = JSON.stringify(d, null, ' ');
			if(v.length <150) v = JSON.stringify(d);
			if( v && v.length > 20000) v = v.substr(0,20000)+'... MORE'; 
			return '<pre>' + (v || (d !== undefined?asHTML(d.toString()):undefined)) + '</pre>';
		}
		else if( d ){
			return prettyPrint(d).outerHTML;
		}
	}

	function smallArray(a, depth=0, len=40) {
		if(depth > 1) return false;
		if(Array.isArray(a) && a.length <= len) {
			if(a.every(isPrimitive)) return true;
			return a.every( v => smallArray(v, depth+1, 3));
		}
		return false;
	}



	function isReactNode(d) {
		if(typeof d !== 'object' || typeof window.React === undefined || typeof window.ReactUI === undefined) return false;
		return d.$$typeof && d.$$typeof.toString() === 'Symbol(react.element)' && !!d.type;
	}

	function isPrimitive(v) {
		switch(typeof v) {
		case 'boolean':
		case 'number': return true;
		case 'string': return v.length < 20;
		}
		return false;
	}

	function __2ToDisplay(title, val) {
		return (`
			<div class="ui container grid">
				<div class="three wide column expression" title="Expression">${title}</div>
				<div class="thirteen wide column expression-value">${display(val)}</div>
			</div>`);
	}

	function pagePrev() { pagePrevNext(this, false); }
	function pageNext() { pagePrevNext(this, true); }

	function pagePrevNext(elem,forward) {
		let curPage = findSegment(elem);
		let targetPage = forward?
			curPage.nextElementSibling:
			curPage.previousElementSibling; 
			
		jumpTag(targetPage,60);
		return [targetPage, curPage];
	}

	function _displayEval(string) {
		if(typeof string === 'string') {
			let title = asHTML(string);
			let val;
			try {
				val = (1,eval)(string);
				if( !(val instanceof Promise) ) pushDisplay(__2ToDisplay(title, val));
				else val.then(val => {
					pushDisplay(__2ToDisplay(title, val));
				});

			} catch(err) {
				pushDisplay('<span class="red">Expression error</span>');
			} // end try
		}
		else _show(string);
	}

	/**
	 * [codeHighlight description]
	 * @param  {[type]} _lines [description]
	 * @return {[type]}        [description]
	 */
	function codeHighlight(_lines) {
		//var HL = escapeHTML;
		var HL = (s,type) => hljs.highlightAuto(s, hljsLang(type)).value; 
		var [list, type, content] = 
					_lines.reduce( ([list, type, content], line) =>{
						let mat = line.match(/^\s*(![a-z_-]+|!--)/);
						if(mat) {
							if(content || type.match(/!render-(start|end)|!head|!end/)) 
								list.push([type, content]);
							return [list, mat[1], ''];
						}

						line = (content?'\n':'')+_indent(line, s => HL(s, type) );
						return [list, type, content+line];
					}, 
					[[], '', '']
					);

		if(content || type === '!end') {
			list.push([type, content]);
		}
				
		if(list.length > 0 && list[0][0] === '!html' && emptyContent(list[0][1]) ) list = list.slice(1);
		return list.flatMap( 
			([type, body]) =>  genTryitSection(type)+`<pre class="${codeBackground(type)}">${body}</pre>`
		).join('<br/>\n')+'<code>&nbsp;<br><br></code>\n';
	}


	function genTryitSection(type) {
		if(!type) return '';
		return `<div class="tryit-section">${type}</div>`; 
	 
	}

	function _indent(s,fn) {
		let i = 0;
		while(s[i] === ' ' || s[i] === '\t') i++;
		return s.substr(0,i) + fn(s.substr(i));
	}



	function codeBackground(type) {
		switch(type) {
		case '!md': return 'code-md';
		case '!tryit': return 'code-tryit';
		case '!html': return 'code-html';
		case '!head': return 'code-head';
		case '!end': return 'code-end';
		}
		return '';
	}

	function emptyContent(str) {
		if(str === undefined) return true;
		for(let i=0; i<str.length; i++) {
			let c = str[i];
			if(c !== ' ' && c != '\t' && c !== '\n' && c !== '\r') return false;
		}
		return true;
	}

	function hljsLang(name) {
		switch(name) {
		case '!md' :  return ['markdown', 'html'];
		case '!head':
		case '!tail':
		case '!html' :  return ['html', 'javascript', 'css'];
		case '!js' :  return ['javascript', 'xml'];
		default :  return undefined;
		}
	}

	function beforeExecute(divName) {
		let $$ = window.$$ || {};
		$$.executeDiv = divName;
		if(typeof $$.beforeExecute === 'function') {
			try {
				$$.beforeExecute(divName);
			} catch (err) {
				console.log(err);
			}
		}
	}

	// import {$e} from '../utils';
	// import { runLastly } from '../display/lastly';
	// import { setNoDisplay } from '../display/getSetNoDisplay';
	// import _runAll from './_runAll';
	// import execute from './execute';
	// import showPopup from '../display/showPopup';



	// export default function tryIt(divName,editor, toDelay=200) {

	// 	if(!canExecute(divName)) {
	// 		setTimeout(() => _runAll(getPendingEditors(), divName, true), 300);
	// 		return;
	// 	}
	// 	var _err = $e(divName + '-error');
	// 	var _disp = $e(divName + '-display');
	// 	_err.style.display = 'none';
	// 	_err.innerHTML = '';
	// 	//_disp.style.display = "none";
	// 	_disp.innerHTML = '';
	// 	_disp.style['max-height'] = '100rem';
	// 	setNoDisplay(false);
	// 	setTimeout( () => execute(divName, editor, true, true, runLastly),toDelay);
	// }


	function canExecute(tag) {
		let ix = getPendingEditors().indexOf(tag);
		if( ix <= 0) return true;
		showPopup(3,'Executing all preceeding code snippet, this may take some time');
		//showPopup(3,() => jump(getPendingEditors()[0]));
		//jump(getPendingEditors()[0]);
		return false;
	}

	function jsxCompiler(s) {
		if(!s) return '';
		if(s.match(/<\/|\/>/)) return jsxLoader.compiler.compile(s);
		return s;
	}

	function render(val, lastExecTime=0) {
		if(arguments.length > 0 ) _show(val);
		let _displayStack = getDisplayStack();
		let promises = _displayStack.map( ([p]) => p);
		let types = _displayStack.map( ([,type]) => type);
		let resPromise = Promise.all(promises).then(
			list => {
				if(!getNoDisplay()) {
					let res = list.map((v,i) => [ v,types[i] ]);
					return Promise.resolve(
						'<div class="ui accordion">'+
									'<div class="active title"><i class="dropdown icon"></i>Results ( '+round2(lastExecTime)+' ms)</div>'+
									'<div class="active content">'+
									res.map( ([v,type]) => type==='h'? v : display(v) ).join('\n') + 
								'</div></div>'
					);
				}
				else return Promise.resolve(undefined);
			}
		);
		clearDisplay();
		return resPromise;
	}

	function round2(time) { return Math.round(time*100.0)/100.0; }

	function replaceCSSClass(tag) {
		let __editorsPending = getPendingEditors();
		let ix = getPendingEditors().indexOf(tag);
		if( ix !== 0) return false;
		__editorsPending = __editorsPending.slice(1); // 
		setPendingEditors(__editorsPending);
		let divName = __editorsPending[0];
		addRemoveCSSclass(divName, 'yellow', 'green').dataset.tooltip = 'Execute Script (Ctrl+Enter)';
		return true;
	}

	function updateUI(divName,toJump=true) {
		replaceCSSClass(divName);
		_addRemoveCSSclass(divName+'-run',['green', 'yellow'], 'blue').dataset.tooltip = 'Re-Execute Script (Ctrl+Enter)';
		if(toJump) setTimeout( () => jump(divName),0);
	}

	function execute(divName, editor, toUpdateUI, toJump, callback) {
		try { 
			//CHANGED = true;
			let t0 = performance.now();
			beforeExecute(divName); 
			let displaySeg = $e(divName + '-display');
			let output = $e(divName + '-output');
			let boundingSeg = output.closest('.tryit-inner');
			boundingSeg.closest('.tryit-inner').style.setProperty('margin-bottom', '-1.9rem');

			output.style.display = 'block';
			jsxLoader.compiler.addUseStrict = false;
			var val = (1,eval)(jsxCompiler(editor.getValue('\n'))); // execute script in global context
						
			let lastExecTime = performance.now()-t0;
						
			let show = val => displaySeg.innerHTML = val;
			displaySeg.style.display = 'block';

			render(val,lastExecTime).then(res => {
				if(res !== undefined) show(res);
				if( toUpdateUI ) updateUI(divName );
				if(callback) callback();
				$('.ui.accordion').accordion() ;
			});

		} catch (e) {
			var err = $e(divName + '-error');
			err.innerText = e.toString()+e.stack.toString();
			err.style.display = 'block';
			console.log(e.stack);
			clearDisplay();
			setTimeout( () => jump(divName),0);
		}
	}

	function progress(action, maxV) {
		const $progress = $('.execution.progress .ui.progress');

		switch(action) {
		case 'init' : {
			//alert('init '+maxV);
			$progress.progress({
				total    : maxV,
				text     : {
					active: '{value} of {total} done'
				}
			});
			$progress.progress('reset');
			$progress.data('value', 1);
			$progress.data('total', maxV);
			$progress.data('tryitdelay', maxV>5 ? 1500: 500);
			$('.execution.progress').css('display','block');
			break;
		}
		case 'step': {
			//alert('step');
			$progress.progress('increment');
			break;
		}
		case 'done': {
			let delay = $progress.data('tryitdelay');
			setTimeout(() => $('.execution.progress').css('display','none'), delay);
		}
		}
	}

	function tryIt$2(divName,editor, toDelay=200) {
		let __editorsPending = getPendingEditors(); 
		if(!canExecute(divName)) {
			setTimeout(() => _runAll(__editorsPending, divName, true), 300);
			return;
		}
		var _err = $e(divName + '-error');
		var _disp = $e(divName + '-display');
		_err.style.display = 'none';
		_err.innerHTML = '';
		//_disp.style.display = "none";
		_disp.innerHTML = '';
		_disp.style['max-height'] = '100rem';
		setNoDisplay(false);
		setTimeout( () => execute(divName, editor, true, true, runLastly),toDelay);
	}

	function _runAll(list, item, toInit) {
		let [divName, ...newList] = list;
		if(toInit) {
			let ix = list.indexOf(item);
			if( ix === -1 ) ix = list.length-1;
		
			progress('init', ix);
		}
		if(item === divName){ 
			let editor = editorFor[divName];
			progress('done');
			setTimeout(() => tryIt$2(divName, editor), 200);
			return;
		}
		let _code;

		try {
			console.log('run all '+divName);
			setNoDisplay(true);
			let editor = editorFor[divName];
			if(!item) {
				console.log('item '+divName+'not found');
				progress('done');
				return;
			}

			beforeExecute(divName);
			jsxLoader.compiler.addUseStrict = false;

			_code = editor.getValue('\n');
			var val = (1,eval)(jsxCompiler(_code)); // execute script in global context(_code);

			render(val).then(() => {
				//replaceCSSClass(divName, false);
				updateUI(divName,false);
				progress('step');
				setTimeout( () => _runAll(newList, item),1);
			}).catch(e => (clearLastly(), progress('done'), alert(e)));

		} catch (e) {
			var err = $e(divName + '-error'); 
			err.innerText = e.toString()+e.stack.toString()+'\n\ndiv:'+divName+'\n-------------------------\n'+asHTML(_code);
			err.style.display = 'block';
			progress('done');
			console.log(e.stack);
			clearDisplay();
			clearLastly();
			setTimeout( () => jump(divName),0);
		}
		

	}

	/**
	 * Revert all user changes to scripts to the original script data
	 * @return {void} returns nothing
	 */
	function revertChanges() {
		Object.keys(editorFor).forEach( id => {
			let originalText = $e(id).value;
			let editor = editorFor[id];
			editor.setValue(originalText);
			let theme = tryit$colors.original;
			editor.setOption('theme', theme);
			editor.tryitState = theme;
		});
	}

	function clearStorage() {
		if (confirm('Are you sure you want to clear saved edits')) {
			delete window.localStorage[WINDOW_LOCATION];
			revertChanges();
			console.log('All saved edits removed');
		} 
	}

	function setEditorValue(id) {
		let editor = editorFor[id];
		let v = editor.getValue('\n');
		let editorData = getEditorData();
		let originalContents = $e(id).value;
		if( v !== originalContents) {
			editorData[id] = { key: id, hash: sha1(originalContents), content: v};
			let theme = tryit$colors.saved;
			editor.setOption('theme', theme);
			editor.tryitState = theme;
		}
		return editorData;      
	}

	/**
	 * Save content of editor named
	 * @param  {[type]} id [description]
	 * @return {[type]}    [description]
	 */
	function save(id) {
		window.localStorage[WINDOW_LOCATION] = JSON.stringify(setEditorValue(id));
	}

	/**
	 * Save data from all the editors
	 * @return {undefined} no return vales
	 */
	function saveAll() {
		setEditorData({});  // clear out 
		Object.keys(editorFor).forEach( id => {
			setEditorValue(id);
		});
		const editorData = getEditorData();
		window.localStorage[WINDOW_LOCATION] = JSON.stringify(editorData);

		alert('Save All');
	}

	function gMakeEditor() {
		getPageInfo();
		// let pageInfo = pi.pageInfo;
		// let allEditors = pi.allEditors;
		qsA('div[data-pagevisible="true"]')
			.forEach(e => setDisplay(e, 'false'));
		//			setDisplay(qs('div[data-pagevisible]'),'true');
			
		(qs('.save_all')||{}).onclick = saveAll;
		(qs('.clear_storage')||{}).onclick = clearStorage;
		(qs('.revert_changes')||{}).onclick = revertChanges;

		qsA('.jump_next')
			.forEach(n => {
				let id = n.id.substr(5); 
				n.onclick=(()=>jump(id));
				n.dataset.tooltip='Jump to next script';
			}
			);
		qsA('.jump_back')
			.forEach(n => {
				n.onclick=jumpback;
				n.dataset.tooltip='Jump to ready to execute script';
			}
			);
		qsA('.run_all')
			.forEach(n => {
				let id = n.id.substr(3); 
				n.onclick=() => _runAll(getPendingEditors(), 'tryit'+id, true);
			}
			);
		qsA('.save_data')
			.forEach(n => {
				let id = n.id.substr(5); 
				n.onclick=() => save('tryit'+id);
				n.dataset.tooltip = 'Save this script';
			}
			);
		//_addRemoveCSSclass('ra_1',"green", "grey").style ="display: none";
	}

	/*
	   Perform custom highlighting for TryitJS code
	 */


	function highlightCodeBlock(block) {
		if(!window.$$) throw new Error('$$ not set up');
		let $$ = window.$$;
		if (!block || !hljs || !window.$$ ) return;

		if (block.classList.contains('language-tryit')) {
			var _lines = (unescape(block.innerText) || '').split('\n');

			block.innerHTML = $$.codeHighlight(_lines);
		} else hljs.highlightBlock(block);
	}


	function unescape(s) {
		return s.replace(/~~lt%%|~~gt%%|~~amp%%|"~~code%%"/g, c => {
			switch(c) {
			case '~~lt%%': return '<';
			case '~~gt%%': return '>';
			case '~~amp%%': return '&';
			case '~~code%%': return '```';
			}
		});
	}

	/**
	 * [showPopup description]
	 * @param  {[type]} timeout [description]
	 * @param  {[type]} msg     [description]
	 * @param  {[type]} type    [description]
	 * @param  {[type]} action  [description]
	 * @return {[type]}         [description]
	 */
	function showPopup$1(timeout, msg,type, action){
		action = action || Identity;
		alertify.notify((msg||'Executing all preceeding code snippet, this may take some time'),(type||'error'),timeout, action );
	}

	function getAllEditors() {
		return getPageInfo().allEditors;
	}

	/**
	 * Check if any editor has unsaved changes
	 * 
	 * @return {boolean} `true` - unsaved changes, `false` otherwise
	 */
	function unsavedChanges() {
		try {
			return getAllEditors().some(e => !e.isClean());
		} catch(e) {
			return true;
		}
	}

	setupPageInfo(tryIt$2); // break circular reference
	setupMakeEditor(tryIt$2);
	window.onbeforeunload = function() { 
		if(unsavedChanges())
			return 'You have made changes on this page that you have not yet confirmed. If you navigate away from this page you will lose your unsaved changes';
	};

	document.addEventListener('DOMContentLoaded', () => {

		// check if we have highlightings then highlight TryitJS code snippets	
		if(hljs) { 
			qsA('pre code.language-tryit').forEach(highlightCodeBlock);
			qsA('pre code.language-js').forEach(highlightCodeBlock);
			qsA('pre code.language-javascript').forEach(highlightCodeBlock);
		}
		gMakeEditor();
		let allPages = qsA('div[data-pagevisible]');

		// show only the first page
		allPages.forEach(
			(elem,i) => 
				i!==0?
					setDisplay(elem,'false'):
					''
		);


		qsA('.page_prev').forEach(e => { e.onclick = pagePrev; e.dataset.tooltip='Go to previous page (Key: 🡄)';} );
		qsA('.page_next').forEach(e => { e.onclick = pageNext; e.dataset.tooltip='Go to next page (Key: 🡆)';} ); 

		$('pre:has(code.language-tryit)').addClass('language-tryit');
		if(location.hash) {
			setTimeout(() =>jumpTag(location.hash.substr(1),false),  0);
		}
		else {
			setTimeout(() =>jumpTag('page-1',false),  0);
		}

	});

	document.addEventListener('keydown', keydown);
	qsA('button').forEach(el => el.addEventListener('keydown', keydown));

	function keydown(event) {
		const LeftArrow = 37, RightArrow = 39;
		const activeElement = document.activeElement;
		if (navigator.platform === 'MacIntel' ? event.metaKey : event.ctrlKey && event.key === 's') {
			event.preventDefault();
			saveAll();
			// ... your code here ...
		}
	   
		else if( (activeElement === document.body || isTag(activeElement, 'button')  ) && 
				(event.keyCode === LeftArrow /*KeyLeft */ || event.keyCode === RightArrow /*key right */)) {
			let keyCode = event.keyCode;
			let p = qs('div.try-page[data-pagevisible=true]');
			let elem =  p && p.querySelector(keyCode==LeftArrow?'.page_prev':'.page_next');
			if(!elem) {
				showPopup$1(1, keyCode==RightArrow?'Last Page':'First Page','success');	
			}
			elem && (event.preventDefault(), elem.onclick());
		} 

	}

	window.addEventListener('popstate', e=> {
		let _hash = e.target.location.hash.substr(1);
		console.log(e);
		if(_hash && LAST_TARGET !== _hash) {
			jumpTag(_hash,20,undefined, true);
		}

	});

	var init = ({LAST_TARGET});

	var $$ = {
		codeHighlight: codeHighlight,
		D:    _show,
		D2:   _displayEval,
		HTML: pushDisplay,
		show: _show,
		clear:  clearDisplay,
		render: render,
		objInfo:objInfo,
		executeDiv: '',     // the tryit div being executed
		beforeExecute: () => init,  // placeholder 
		H:H,
		lastly: _lastly, // pass a function after all items have been displayed, this call be called several
		// times, the actions are performed in the order they are posted
		json: (...v) => _show(...v.map(json)),
		prettyPrint
	};



	window.$$ = $$;
	const toExport = ({
		gMakeEditor,
		$$,              //display interface
		jumpTag: _jumpTag,
		jumpBack,
		_display: display,
		getPendingEditors,
		pagePrev,
		pageNext,
		asArray,
		saveAll,
		qs,
		qsA,
		//pageVisibleBefore,
		showPopup: showPopup$1,
		unsavedChanges,
		H,
		escapeHTML: asHTML
	});

	Object.keys(toExport).forEach(k => window[k] = toExport[k]);

	return toExport;

}());
